/***************************************************************************
 *  Copyright (C) 2003 Daniel Franke                                       *
 *                                                                         *
 *  Authors:                                                               *
 *      Daniel Franke <daniel.franke@imbs.uni-luebeck.de>                  *
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 ***************************************************************************/

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <cctype>
#include <algorithm>
#include <functional>
#include <iterator>
#include <list>
#include <cerrno>
using namespace std;

#include "db.h"

#define PACKAGE_STRING "abi2link 1.0.0"
#define PACKAGE_BUGREPORT "<daniel.franke@imbs.uni-luebeck.de>"

ofstream abilog;
string __G_prefix;
bool   __G_estimateFromFounderOnly;	// true := estimate==founder, false := estimate==all

const string missingValuePhenotype = "-999999";
// const string missingValueGenotype  = "0";


static void version(void) {
        cout << PACKAGE_STRING << endl
             << "Written by Daniel Franke." << endl
             << endl
             << "Copyright 2003 by Daniel Franke." << endl
             << "This is free software; see the source for "
             << "copying conditions. There is NO" << endl
             << "Warranty; not even for MERCHANTABILITY "
             << "or FITNESS FOR A PARTICULAR PURPOSE." << endl;

        exit(EXIT_SUCCESS);
}

static void usage(const char *progname) {
        cout << PACKAGE_STRING << endl
             << "Written by Daniel Franke." << endl
             << endl
             << "Usage: " << progname  << " ARGUMENTS" << endl
	     << endl
	     << "Currently known arguments:" << endl
	     << "        --map <haldane|kosambi>" << endl
	     << "                         locus mapping function" << endl
	     << "        --ped <file>     pedigree file" << endl
	     << "        --chr <file>     chromosome description file" << endl
	     << "        --trait <file>   trait file (optional)" << endl
	     << endl
	     << "        --estimate <all|founder>" << endl
	     << "                         whether to estimate allel frequencies from all" << endl
	     << "                         individuals or founders only (optional, default: all)" << endl
	     << "        --prefix <name>  output file prefix (optional, default: abi2link)" << endl
             << endl
             << "    -v, --version        print version information and exit"
             << endl
             << "    -h, --help           print this text and exit" << endl
	     << endl
	     << "Report errors to " << PACKAGE_BUGREPORT << endl;

        exit(EXIT_SUCCESS);
}


// One should not inherit from ifstream but take an ifstream as one
// of the constructor's arguments
class abistream : public ifstream {
public:
	abistream(const char *f) : ifstream(f),
	    abifile(f), nWarning(0), nError(0) {
		if (! good()) {
			abilog << f << ": error: " << strerror(errno) << endl;
			nError++;
		}
	}

	abistream(string f) : ifstream(f.c_str()),
	    abifile(f), nWarning(0), nError(0) {
		if (! good()) {
			abilog << f << ": error: " << strerror(errno) << endl;
			nError++;
		}
	}

	virtual ~abistream(void) { }

	abistream& preprocess(void);

	void parse(void) {
		// `std::for_each' doesn't seem to work here
		// maybe the 'UnaryFunction' of 'for_each' has to built up with
		// 'std::compose1' or similar?
		// for_each(lines.begin(), lines.end(), compose1());

		list<string>::const_iterator it = lines.begin();
		while(it != lines.end())
			parseLine(*it++);
	}

	virtual void postprocess(void) = 0;

	virtual void report(void) const {
		cout << __G_prefix << ".log: there were "
		     << nWarning << " warning(s) and "
		     << nError << " error(s) in "
		     << abifile
		     << endl;
	}

private:
	abistream& skipws(void);
	abistream& skipcomment(void);

protected:
	virtual void parseLine(const string& line) = 0;

	string abifile;
	list<string> lines;

	unsigned int nWarning, nError;
};

abistream& abistream::preprocess(void) {
	//
	// - remove comments: '*'
	// - remove leading and trailing whitespaces
	// - condense whitespaces between data columns
	//

	if (! good())
		return *this;

	char c;
	string str;
	while(! eof()) {
		c = get();
		while (isalnum(c) || (ispunct(c) && c != '*')) {
			str += c;
			c = get();
		}

		if (c == ' ' || c == '\t') {
			skipws();
			str += ' ';
		} else if (c == '*') {
			skipcomment();
		} else if (c == widen('\n'))
			if (! str.empty()) {
				lines.push_back(str);
				str.clear();
			}
	}

	// write preprocessed lines to stdout
//	copy(lines.begin(), lines.end(), ostream_iterator<string>(cout, "\n"));

	return *this;
}


abistream& abistream::skipws(void) {
	char c = get();
	while (c == ' ' || c == '\t')
		c = get();

	unget();	// ups, one too far
	return *this;
}

abistream& abistream::skipcomment(void) {
	char c = get();
	while (c != widen('\n'))
		c = get();

	unget();	// ups, one too far
	return *this;
}

// implement parsing of pedigree files
class abistreamPed : public abistream {
public:
	abistreamPed(const char *f, bool ignore = false)
		: abistream(f), cnt(0), nAff(-1), ignoreAffectionStatus(ignore) {}

protected:
	void parseLine(const string& line) {
		//
		// multiple columns expected
		//  1. pedigree_id
		//  2. person_id
		//  3. father_pid
		//  4. mother_pid
		//  5. sex
		//  6. affection_status* (optional, none, one or many)
		//

		//
		// One could use:
		//
		// 	list<string> cols(istream_iterator<string>(line),
		//			  istream_iterator<string>());
		//
		// to read in all columns. This would result in code
		// not too intuitive in later usage. Therefore use
		// named variables instead of a list of strings.
		//
		istringstream pedstream(line);
		string ped, pid, father, mother, sex ;

		pedstream >> ped >> pid >> father >> mother >> sex;

		// read as many values as there might be, assuming
		// affection status and liability classes
		string aff;
		list<string> afflist;
		while( pedstream >> aff )
			afflist.push_back(aff);

		// build affection string
		list<string>::iterator it = afflist.begin();
		if (it != afflist.end()) {
			aff = *it++;
			while (it != afflist.end())
				aff = aff + " " + *it++;
		}

		if (nAff == -1)
			nAff = afflist.size();
		else if (nAff != (signed)afflist.size()) {
			abilog << abifile
			     << ": error: individual "
			     << pid
			     << ", number of liability classes differ (here: "
			     << afflist.size() << ", before: " << nAff << ")"
			     << endl;

			     nError++;
		}

		if (pid.empty()) {
			abilog << abifile
			     << ": warning: skipping truncated line '"
			     << line << "'"
			     << endl;
			     nWarning++;
			return;
		}

		if (father.empty()) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing father id set to '0'"
			     << endl;
			father = "0";
			nWarning++;
		}

		if (mother.empty()) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing mother id set to '0'"
			     << endl;
			mother = "0";
			nWarning++;
		}

		if (sex.empty()) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing sex set to '0' (unknown)"
			     << endl;
			sex = "0";
			nWarning++;
		}

		//
		// Daniel Franke, 2003/09/30:
		//	traitfile may be omitted: 'crimap' compatible output
		//	(therefore also no affection status allowed!)
		//
		if (ignoreAffectionStatus && !aff.empty()) {
			abilog << abifile
			       << ": warning: individual " << pid
			       << ", affection status ignored (trait undefined)"
			       << endl;
			aff.clear();
			nWarning++;
		}

		if (aff.empty() && !ignoreAffectionStatus) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing affection status set to '2'"
			     << endl;
			aff = "2"; // Andre: set undefined affstat to 'affected'
			nWarning++;
		}

		//
		// Before adding a new person, check if it was
		// defined already - if yes discard the current data
		// and inform user about this error.
		// If the current person is not available, add it to the db.
		//
		// I'm not sure if checks like this shouldn't be done in
		// 'postprocess()'
		//

		// get current persons (pid should be unique)
		PersonTable pTable = db.db_person.selectWherePID(pid);

		if (! pTable.empty()) {

			abilog << abifile << ":error : individual "
			     << pid
			     << " defined multiple times"
			     << endl;
			nError++;
			// discard data silently
			return;
		}

		// create database entry
		struct Person *p  = new struct Person;
		p->prsn_ped       = ped;
		p->prsn_pid       = pid;
		p->prsn_fatherpid = father;
		p->prsn_motherpid = mother;
		p->prsn_sex       = sex;
		p->prsn_affected  = aff;

		db.db_person.append(++cnt, p);
	}

	void postprocess(void) {
		// nothing to do (?)
	}

private:
	mutable int cnt;
	int nAff;			// number of columns of affection status
	bool ignoreAffectionStatus;
};


// implement parsing of trait files
class abistreamTrait : public abistream {
public:
	abistreamTrait(const char *f) : abistream(f) {}
	void parseLine(const string& line) {
		//
		// two columns expected:
		//  1. person_id
		//  2. trait_value
		//

		istringstream traitstream(line);
		string pid, trait;

		traitstream >> pid >> trait;

		if (trait.empty()) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing trait set to '" << missingValuePhenotype << "'"
			     << endl;

			trait = missingValuePhenotype;
			nWarning++;
		}

		//
		// Before adding a new trait, check if the corresponding
		// person already had one assigned -
		// if yes discard the current data and warn user.
		// Otherwise add it to the db.
		//
		// I'm not sure if checks like this shouldn't be done in
		// 'postprocess()'
		//

		// get current phenotypes (pid should be unique)
		PhenotypeTable phenoTable = db.db_phenotype.selectWherePID(pid);

		if (! phenoTable.empty()) {

			abilog << abifile << ":error : individual "
			     << pid
			     << " has multiple traits assigned"
			     << endl;

			nError++;
			// discard data silently
			return;
		}


		// create database entry
		struct Phenotype *p = new struct Phenotype;
		p->pheno_pid   = pid;
		p->pheno_value = trait;

		db.db_phenotype.append(pid, p);
	}

	void postprocess(void) {
		// check whether there trait values without persons
		PhenotypeTable::iterator phenoit = db.db_phenotype.begin();
		for ( ; phenoit != db.db_phenotype.end(); ++phenoit) {
			if (db.db_person.selectWherePID((*phenoit).first).empty()) {
				abilog << abifile << ": warning: individual "
				     << (*phenoit).first
				     << " has trait value"
				     << ", but not in pedigree"
				     << endl;
				nWarning++;
			}
		}

		// check whether there are persons without trait
		PersonTable::iterator pit = db.db_person.begin();
		for ( ; pit != db.db_person.end(); ++pit) {
			IdType pid = (*pit).second->prsn_pid;

			if (db.db_phenotype.selectWherePID(pid).empty()) {
				abilog << abifile << ": warning: individual "
				     << pid
				     << ", missing trait set to '" << missingValuePhenotype << "'"
				     << endl;
				nWarning++;

				struct Phenotype *p = new struct Phenotype;
				p->pheno_pid        = pid;
				p->pheno_value      = missingValuePhenotype;

				db.db_phenotype.append(pid, p);
			}
		}
	}
};

// implement parsing of genotype files
class abistreamGen : public abistream {
public:
	abistreamGen(const char *l, const char *f)
		: abistream(f), locusname(l) { }

	void parseLine(const string& line) {
		//
		// three columns expected:
		//  1. person_id
		//  2. genotype_1
		//  3. genotype_2
		//
		istringstream genstream(line);
		string pid, gen1str, gen2str;

		genstream >> pid >> gen1str >> gen2str;

		AllelType gen1, gen2;

		if (gen1str.empty()) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing genotype set to '0'"
			     << endl;

			gen1 = 0;
			nWarning++;
		} else {
                        char *e = NULL;
                        gen1 = strtol(gen1str.c_str(), &e, 10);
                        if (*e) {
                                abilog << abifile
                                     << ": error: individual " << pid
                                     << ", non-integer allelotype '"
				     << gen1str << "' (set to '0')"
                                     << endl;
                                gen2 = 0;
				nWarning++;
                        }

		}

		if (gen2str.empty()) {
			abilog << abifile
			     << ": warning: individual " << pid
			     << ", missing genotype set to '0'"
			     << endl;

			gen2 = 0;
			nWarning++;
		} else {
			char *e = NULL;
			gen2 = strtol(gen2str.c_str(), &e, 10);
			if (*e) {
				abilog << abifile
				     << ": error: individual " << pid
				     << ", non-integer allelotype '"
				     << gen2str << "' (set to '0')"
				     << endl;
				gen2 = 0;
				nWarning++;
			}
		}

		//
		// Before adding a new genotype, check if it was
		// defined already - if yes, and genotypes are the same
		// omit the current genotype and warn the user about it
		// When genotypes differ it should be considered to be
		// an error condition.
		// If the current locus is not available, add it to the db.
		//
		// I'm not sure if checks like this shouldn't be done in
		// 'postprocess()'
		//
		// At no time there should be more than one genotype per
		// locus and person in 'db.db_genotype' .
		// The code below relies on this assumption.
		//

		// get (all) genotypes of current person
		GenotypeTable prsnGenTable = db.db_genotype.selectWherePID(pid);
		// get genotypes of current locus from all genotypes
                GenotypeTable genTable = prsnGenTable.selectWhereLocusID(locusname);

		if (! genTable.empty()) {
			struct Genotype *g = (*genTable.begin()).second;

			// check if genotypes differ
			if ((g->geno_paternal == gen1 && g->geno_maternal == gen2) ||
			    (g->geno_paternal == gen2 && g->geno_maternal == gen1)) {
				abilog << abifile << ": warning: discarded multiple "
				     << " occurence of genotype ("
				     << gen1 << ", " << gen2
				     << ") of individual "
				     << pid
				     << endl;
				     nWarning++;
			} else {
				abilog << abifile << ": error: discarded different "
				     << " genotypes ("
				     << g->geno_paternal << ", " << g->geno_maternal
				     << ") and ("
				     << gen1 << ", " << gen2
				     << ") of individual "
				     << pid
				     << endl;
				     nError++;
			}

			// whatever condition, DO NOT ADD the current genotype info
			return;
		}

		// add new genotype for this person at current locus
		struct Genotype *g = new struct Genotype;
		g->geno_pid        = pid;
		g->geno_locid      = locusname;
		g->geno_paternal   = gen1;
		g->geno_maternal   = gen2;

		db.db_genotype.append(pid, g);
	}

	void postprocess(void) {
		// check whether there are genotypes without persons
		GenotypeTable::iterator genoit = db.db_genotype.begin();
		for ( ; genoit != db.db_genotype.end(); ++genoit) {
			IdType pid = (*genoit).first;

	                if (db.db_person.selectWherePID(pid).empty()) {
	                        abilog << abifile << ": warning: individual "
        	                     << pid
                	             << " genotyped, but not in pedigree"
                	             << endl;
				nWarning++;
               		}
		}

		//
		// check whether there are persons without genotypes
		// (at current locus 'locusname')
		//
		// There aren't multiple genotypes, 'parseLine()' takes care of this.
		//
		PersonTable::iterator pit = db.db_person.begin();
		for ( ; pit != db.db_person.end(); ++pit) {
			IdType pid = (*pit).second->prsn_pid;

			// get (all) genotypes of current person
			GenotypeTable prsnGenTable = db.db_genotype.selectWherePID(pid);
			// get genotypes of current locus from all genotypes
	                GenotypeTable genTable = prsnGenTable.selectWhereLocusID(locusname);
			if (genTable.empty()) {
				abilog << abifile << ": warning: individual "
				     << pid
	                             << " not genotyped, set to '0'"
	                             << endl;
				nWarning++;

				struct Genotype *g = new struct Genotype;
				g->geno_pid         = pid;
				g->geno_locid       = locusname;
				g->geno_paternal    = 0;
				g->geno_maternal    = 0;

				db.db_genotype.append(pid, g);
	                }
		}

		//
		// due to limitation of genetic analysis software
		// we have to recode alleles, the the smallest value
		// maps to 1 and so on
		//   13 -> 1
		//  127 -> 2
		//  130 -> 3
		//

		// get the genotypes of all individuals
		GenotypeTable genTable = db.db_genotype.selectWhereLocusID(locusname);
		GenotypeTable::iterator git = genTable.begin();

		// build up a list
		map<AllelType, AllelType> allelmap;
		for ( ; git != genTable.end(); ++git) {
			struct Genotype *g = (*git).second;
			if (allelmap.find(g->geno_paternal) == allelmap.end())
				allelmap[g->geno_paternal] = 0;
			if (allelmap.find(g->geno_maternal) == allelmap.end())
				allelmap[g->geno_maternal] = 0;
		}

		// number
		int n = (*allelmap.begin()).first == 0 ? 0 : 1;
		for (map<AllelType, AllelType>::iterator ait = allelmap.begin();
		     ait != allelmap.end(); ++ait)
		     	(*ait).second = n++;

		// recode
		for (git = genTable.begin(); git != genTable.end(); ++git) {
			struct Genotype *g = (*git).second;

			g->geno_paternal = allelmap[g->geno_paternal];
			g->geno_maternal = allelmap[g->geno_maternal];
		}
	}

protected:
	string locusname;
};

// implement parsing of map files
class abistreamMap : public abistream {
public:
	abistreamMap(const char *f) : abistream(f) {}

	virtual void parseLine(const string& line) {
		//
		// two coloumns expected
		//  1. filename of genotype file (basename used as marker name)
		//  2. locus position in centimorgan
		//

		istringstream mapstream(line);
		string genfilename, genpos;

		mapstream >> genfilename >> genpos;

		// extract marker name from filename
		string locusname = genfilename.substr(0, genfilename.find("."));

		// create database entry
		double position = strtod(genpos.c_str(), NULL);
		struct Locus *locus = new struct Locus;
		locus->loc_id   = locusname;
		locus->loc_pos  = position;

		db.db_locus.append(position, locus);

		//
		// does not belong to parsing stage, but what else to do?
		// v1.0.1 - files should be added to 'files' object in main() ?
		//
		// read genotype data
		//
		abistreamGen gen(locusname.c_str(), genfilename.c_str());
		gen.preprocess();
		gen.parse();
		// here, parsing of persons has to be complete!
		gen.postprocess();
		gen.report();
		gen.close();
	}

	void postprocess(void) {
		// nothing to do (?)
	}
};


// defined in linkage.cpp
extern double haldane(const double& p1, const double& p2);
extern double kosambi(const double& p1, const double& p2);
extern double (*mapfunc)(const double&, const double&);

ostream& linkage_dat(ostream& stream, bool, bool);
ostream& linkage_pre(ostream &out, struct Person *p);

static int extractArg(int argc, char **argv, const char *argname, char **argdata, bool optional=false) {

	char **tmp;
	if((tmp = find(argv, argv + argc, string(argname))) == argv + argc) {
		if  (! optional) {
			cerr << argv[0] << ": argument '" << argname << "' omitted." << endl
			     << "Try '" << argv[0] << " --help' for more information."
			     << endl;
		}

		*argdata = NULL;
		return -1;

	} else if (tmp+1 == argv + argc) {
		if (! optional) {
			cerr << argv[0] << ": argument '" << argname << "' omitted." << endl
			     << "Try '" << argv[0] << " --help' for more information."
			     << endl;
		}

		*argdata = NULL;
		return -1;
	}

	*argdata = *(tmp+1);
	return 0;
}

int main(int argc, char **argv) {

        // if '-v' or '--version' given, print version information, exit
        if((find(argv, argv + argc, string("-v")) != argv + argc) ||
           (find(argv, argv + argc, string("--version")) != argv + argc))
                version();

        // if '-h' or '--help' given, print help text, exit
        if((find(argv, argv + argc, string("-h")) != argv + argc) ||
           (find(argv, argv + argc, string("--help")) != argv + argc))
                usage(argv[0]);


	char *mapfuncname;
	if (extractArg(argc, argv, "--map", &mapfuncname) == 0) {
		if (! string(mapfuncname).compare("haldane"))
			mapfunc = haldane;
		else if (! string(mapfuncname).compare("kosambi"))
			mapfunc = kosambi;
		else {
			cerr << argv[0] << ": error: unknown mapping function '"
			     << mapfuncname << "'" << endl;

			return EXIT_FAILURE;
		}
	} else
		return EXIT_FAILURE;

	list<abistream*> files;

	//
	// Daniel Franke, 2003/09/30:
	//	traitfile may be omitted: 'crimap' compatible output
	//
	bool ignoreAffectionStatus = false;
	bool ignoreTrait = false;

	char *traitfilename;
	if (extractArg(argc, argv, "--trait", &traitfilename, true) == 0)
		files.push_back(new abistreamTrait(traitfilename));
	else
		ignoreAffectionStatus = ignoreTrait = true;

	char *pedfilename;
	if (extractArg(argc, argv, "--ped", &pedfilename) == 0)
		files.push_back(new abistreamPed(pedfilename, ignoreAffectionStatus));
	else
		return EXIT_FAILURE;

	char *chrfilename;
	if (extractArg(argc, argv, "--chr", &chrfilename) == 0)
		files.push_back(new abistreamMap(chrfilename));
	else
		return EXIT_FAILURE;

	char *prefix;
	if (extractArg(argc, argv, "--prefix", &prefix, true) == 0)
		__G_prefix = prefix;
	else
		__G_prefix = "abi2link";

	abilog.open((__G_prefix + ".log").c_str());

	char *estimate;
	if (extractArg(argc, argv, "--est", &estimate, true) == 0) {
		if (string(estimate).compare("founder") == 0)
			__G_estimateFromFounderOnly = true;
		else if (string(estimate).compare("all") == 0)
			__G_estimateFromFounderOnly = false;
		else
			cerr << argv[0] << ": error: unknown allel frequency estimation range '"
			     << estimate << "'" << endl;
	} else
		__G_estimateFromFounderOnly = false;


	// read input
	for_each(files.begin(), files.end(), mem_fun(&abistream::preprocess));
	for_each(files.begin(), files.end(), mem_fun(&abistream::parse));
	for_each(files.begin(), files.end(), mem_fun(&abistream::postprocess));
	for_each(files.begin(), files.end(), mem_fun(&abistream::report));

	// housekeeping: close files, free memory
	list<abistream*>::iterator it = files.begin();
	while (it != files.end()) {
		(*it)->close();
		delete (*it++);
	}
	files.clear();

	// write output
	// the following should somehow also work with something like:
	// 	for_each(db.db_person.begin(), db.db_person.end(),
	//		bind1st(linkage, cout));
	ofstream pre((__G_prefix + ".pre").c_str(), ios::out);
	if (! pre.good()) {
		cerr << __G_prefix << ".pre: error: "
		     << strerror(errno)
		     << endl;
	} else {
		PersonTable::iterator pit = db.db_person.begin();
		while (pit != db.db_person.end()) {
			linkage_pre(pre, (*pit++).second);
		}
		pre.close();
	}

	ofstream dat((__G_prefix + ".dat").c_str(), ios::out);
	if (! dat.good()) {
		cerr << __G_prefix << ".dat: error: "
		     << strerror(errno)
		     << endl;
	} else {
		// ignoreXXX set above
		linkage_dat(dat, ignoreAffectionStatus, ignoreTrait);
		dat.close();
	}


	return 0;
}

