/***************************************************************************
 *  Copyright (C) 2003 Daniel Franke                                       *
 *                                                                         *
 *  Authors:                                                               *
 *   	Daniel Franke <daniel.franke@imbs.uni-luebeck.de>                  *
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 ***************************************************************************/

#ifndef __DB_H__
#define __DB_H__

#include <map>
#include <string>

using std::multimap;
using std::string;


// typedef string StorageType;

class StorageType : public string {
public:
	StorageType(string s = "") : string(s) {}

	inline bool operator()(const StorageType& first, 
			       const StorageType& second) const {
		return first < second;
	}
};

typedef StorageType IdType;
typedef unsigned int AllelType;

/**
 * @brief A table-like template class.
 *
 * The table class layers a STL-multimap object. 
 * The idea: to have an easy to use
 * (and minimal) base class interface. From this base class specialized tables
 * will be derived which implement specialized, SQL-like functions, e.g. as
 * selecting recordsets by given id's. Just 'typenaming' a multimap would
 * have been sufficient in some ways, but this approach is somewhat cleaner,
 * at least, I believe so (all functions not needed are masked).
 *
 * Template parameters:
 * <ul>
 *    <li>key_type - a not necessessarily unique key
 *    <li>record_type - a record type, in most a cases a pointer to 'struct foo'
 *    <li>compare - STL functor (less compareable)
 * </ul>
 *
 * If the record_type is defined as pointer to 'struct foo',
 * compare should be
 * an 'foo::operator(const key_type&, const key_type&)' functor.
 */
template <class key_type, 
	  class record_type, 
	  class compare = std::less<key_type> > 
class table {
public:

	typedef typename std::multimap<key_type, record_type, compare> table_type;
	typedef typename table_type::value_type value_type;
	typedef typename table_type::iterator iterator;
	typedef typename table_type::const_iterator const_iterator;

	/** @brief Constructor, basic initializations. */
	table(void) : mTable() { }

	/** @brief Destructor. */
	virtual ~table(void) { }

	/**
	 * @brief Append record to table.
	 *
	 * Records will be sorted in ascending order corresponding 
	 * to the compare functor.
	 *
	 * @param key	 Key value, used for sorting the record in.
	 * @param record The record to insert.
	 */
	inline void append(const key_type &key, const record_type &record) {
		mTable.insert(value_type(key, record));
	}

	inline void append(const value_type& p) {
		mTable.insert(p);
	}

	/**
	 * @brief Get number of records.
	 *
	 * This method is layered from multimap.
	 *
	 * @return Number of records in table.
	 */
	inline unsigned long size(void) const {
		return mTable.size();
	}

	/**
	 * @brief Determine wether the table is empty or not.
	 *
	 * This method is layered from multimap.
	 *
	 * @return Boolean value wether the table is empty or not.
	 */
	inline bool empty(void) const {
		return mTable.empty();
	}

	/**
	 * @brief Returns a read/write iterator
	 */
	inline iterator begin(void) {
		return mTable.begin();
	}
	inline iterator end(void) {
		return mTable.end();
	}

	/**
	 * @brief Returns a read-only iterator
	 */
	inline const_iterator begin(void) const {
		return mTable.begin();
	}
	inline const_iterator end(void) const {
		return mTable.end();
	}

	/**
	 * @brief Erases all elements in a table.
	 *
	 * Note that this function only erases the elements, and that if the
	 * elements themselves are pointers, the pointed-to memory is not
	 * touched in any way. Managing the pointer is the user's
	 * responsibilty.
	 * This method is layered from multimap.
	 */
	virtual void clear(void) {
		mTable.clear();
	}

protected:
	table_type mTable;
};



struct Frequency {
	IdType    freq_locid;
	AllelType freq_allel;
	double    freq_probability;
};

class FrequencyTable : public table<const IdType, struct Frequency *, IdType> {
public:
	/**
	 * @brief Select all frequency definitions of one locus.
	 *
	 * @param locid locus id
	 * @return A table of frequencies of the given locus.
	 *         An empty table if the locus was not found.
	 */
	FrequencyTable selectWhereLocusID(const IdType& locid);
};

/** @brief Recordset type for genotypes.  */
struct Genotype {
	IdType		geno_pid;
	IdType		geno_locid;
	AllelType	geno_paternal;
	AllelType	geno_maternal;
};

/** @brief Holds recordsets of type 'struct Genotype *'.  */
class GenotypeTable : public table<const IdType, struct Genotype *, IdType> {
public:
	/**
	 * @brief Select a subtable of genotypes of a specified pid.
	 *
	 * @param pid person-id, unique, primary key in GenotypeTable.
	 * @return Either an empty table, if pid not found, or a table
	 *         containing all genotypes of that pid.
	 */
	GenotypeTable selectWherePID(const IdType& pid);
	
	/**
	 * @brief Select a subtable of a specified locus.
	 *
	 * @param locid A locus id.
	 * @return Either an empty table, if locus id is not found, or a table
	 *         containing all genotypes at that locus.
	 */
	GenotypeTable selectWhereLocusID(const IdType& locid);

	/** 
	 * @brief Remove all records, free the allocated memory.
	 */
	void clear(void);
};

/** @brief Recordset type for loci. */
struct Locus {
	IdType loc_id;
	double loc_pos;
};

/**
 * @brief Holds recordsets of type 'struct Locus *'.
 *
 * LocusTable is sorted by locus position (ascending).
 */
class LocusTable : public table<double, struct Locus *> {
public:
	/**
	 * @brief Select a specified locus id.
	 *
	 * Since Locus id SHOULD be unique, one could also return a single
	 * Locus record, but in sake of consistency a LocusTable is
	 * returned
	 *
	 * @param id A locus id.
	 * @return Either an empty table, if locus not found OR if more than one
	 *         locus has the same id, otherwise a table containing one
	 *         record.
	 */
	LocusTable selectWhereID(const IdType& id);
};

/** @brief Recordset type for persons. */
struct Person {
	IdType prsn_ped;
	IdType prsn_pid;
	IdType prsn_fatherpid;
	IdType prsn_motherpid;
	IdType prsn_sex;
	IdType prsn_affected;
};

/** @brief Holds recordsets of type 'struct Person *'. */
class PersonTable : public table<unsigned int, struct Person *> {
public:
	PersonTable selectWherePID(const IdType&);
};

/** @brief Recordset type for phenotypes.  */
struct Phenotype {
	IdType	pheno_pid;
	IdType	pheno_value;
};

/** @brief Holds recordsets of type 'struct Phenotype *'.  */
class PhenotypeTable : public table<const IdType, struct Phenotype *, IdType> {
public:
	/**
	 * @brief select all phenotypes of a specified pid.
	 *
	 * @param pid person id, unique primary key in PhenotypeTable.
	 * @return Either an empty table, if pid not found, or a table
	 *         containing all phenotype records of that pid.
	 */
	PhenotypeTable selectWherePID(const IdType& pid);

	/** 
	 * @brief Remove all records, free the allocated memory.
	 */
	void clear(void);
};


struct db {
	PersonTable		db_person;

	LocusTable		db_locus;
	FrequencyTable		db_frequency;

	GenotypeTable		db_genotype;
	PhenotypeTable		db_phenotype;
};

extern struct db db;		// defined in db.cpp

#endif // !__DB_H__

