/***************************************************************************
 *  Copyright (C) 2003 Daniel Franke                                       *
 *                                                                         *
 *  Authors:                                                               *
 *   	Daniel Franke <daniel.franke@imbs.uni-luebeck.de>                  *
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 ***************************************************************************/

#include <iostream>
#include <fstream>			// class ofstream
#include <iomanip>			// stream manipulation
#include <iterator>
#include <stdexcept>

using namespace std;

#include "db.h"

// defined in abi2linkage.cpp
extern const string missingValuePhenotype;
extern bool __G_estimateFromFounderOnly;	// true := estimate==founder, false := estimate==all


double haldane(const double& p1, const double& p2) {
	return (1 - exp(-2 * fabs(double((p2-p1)/double(100.0))))) / 2.0;
}

double kosambi(const double& p1, const double& p2) {
	return tanh(2 * double((p2-p1)/double(100.0))) / 2.0;
}

// pointer to mapping function
double (*mapfunc)(const double&, const double&) = NULL;


ostream& linkage_pre(ostream &out, struct Person *p) {

	out <<        p->prsn_ped
	    << " " << p->prsn_pid
	    << " " << p->prsn_fatherpid
	    << " " << p->prsn_motherpid
	    << " " << p->prsn_sex;

	if (! p->prsn_affected.empty())
		out << " " << p->prsn_affected;

	//
	// Daniel Franke, 2003/09/21
	//   get genotypes ordered by locus, print warning if either no
	//   or multiple definitions are available
	//
	// Daniel Franke, 2003/09/24
	//   Checks for this are already implemented in 'abistream::postprocess'
	//   There it is taken care that here will be exactly one genotype per
	//   locus and person be available
	//
	GenotypeTable prsnGenTable = db.db_genotype.selectWherePID(p->prsn_pid);

	LocusTable::iterator locus = db.db_locus.begin();
	while (locus != db.db_locus.end()) {
			struct Genotype *g = (*prsnGenTable.selectWhereLocusID((*locus).second->loc_id).begin()).second;
			out << " " << g->geno_paternal
			    << " " << g->geno_maternal;

		locus++;
	}

	//
	// get phenotypes in any order (?)
	// print warning if no trait is available
	//
	// if trait value was missing in file, it was set to 'missingValuePhenotype'
	//
	PhenotypeTable phenoTable = db.db_phenotype.selectWherePID(p->prsn_pid);
	PhenotypeTable::iterator phit = phenoTable.begin();
	while (phit != phenoTable.end())
		out << " " << (*phit++).second->pheno_value;

	out << endl;

	return out;
}


/****************************************************************************/
static void countAllelesAt(struct Locus *loc) {
	map<AllelType, int> alleles;

	GenotypeTable gTable = db.db_genotype.selectWhereLocusID(loc->loc_id);

	GenotypeTable::iterator git = gTable.begin();
	while (git != gTable.end()) {

		//
		// __G_estimateFromFounderOnly == true
		//	--> estimate frequencies from founders only
		// __G_estimateFromFounderOnly == false
		//	--> estimate frequencies from all individuals
		//
		if (__G_estimateFromFounderOnly) {
			//
			// check whether current genotype corresponds
			// to any founder or not
			//
			PersonTable::iterator pit = db.db_person.selectWherePID(git->second->geno_pid).begin();
			if (pit->second->prsn_fatherpid != pit->second->prsn_motherpid)
				continue;	// non-founder, continue
		}

		alleles[(*git).second->geno_paternal]++;
		alleles[(*git).second->geno_maternal]++;
		++git;
	}

	//
	// Note: Allele frequencies will be incorrect if at least
	//       one individual was genotyped multiple times!
	//       User has been notified already.
	//
	// Note: in 'abistream::postprocess' was taken care that
	//       there aren't multiple genotypes in db. Nevertheless,
	//       frequencies might be incorrect here
	//

	// skip '0' genotypes
	map<AllelType, int>::iterator it = alleles.begin();
	if ((*it).first == 0)
		alleles.erase(it);

	// sum number of alleles
	int sum = 0;
	for (it = alleles.begin(); it != alleles.end(); ++it)
		sum += (*it).second;

	// create database entries
	// if 'sum == 0' (division by zero condition) the list of
	// alleles will be empty.
	for (it = alleles.begin(); it != alleles.end(); ++it) {
		struct Frequency *freq = new struct Frequency;
		freq->freq_locid       = loc->loc_id;
		freq->freq_allel       = (*it).first;
		freq->freq_probability = (*it).second / (double)sum;

		db.db_frequency.append(loc->loc_id, freq);
	}
}

static ostream& datAllelNumbers(ostream& stream, struct Locus *loc) {

	countAllelesAt(loc);
	FrequencyTable freqTable = db.db_frequency.selectWhereLocusID(loc->loc_id);

	// First line:
	//  - locus type		3, AllelNumbers
	//  - number of alleles
	//  - locus name
	//  - additional comments
	stream << 3
	       << setw(4) << freqTable.size()
	       << " # " << loc->loc_id
	       << " << ALLELE NUMBERS, NO. OF ALLELES"
	       << endl;

	// Second line:
	//  - gene frequencies
	stream << setiosflags(ios::fixed) << setprecision(5);

	FrequencyTable::iterator fit = freqTable.begin();
	while(fit != freqTable.end())
		stream << setw(9) << (*fit++).second->freq_probability;

	stream << resetiosflags(ios::fixed);
	stream << " << GENE FREQUENCIES" << endl;

	return stream;
}

static ostream& datAffectionStatus(ostream& stream, const struct Locus *loc) {

	// DUMMY FUNCTION - NO REAL DATA!

	// First line:
	//  - locus type		1, AffectionStatus
	//  - number of alleles
	//  - additional comments
	stream << 1
	       << setw(4) << 2
	       << " # disease << AFFECTION, NO. OF ALLELES"
	       << endl;

	// Second line:
	//  - gene frequencies
	stream << setiosflags(ios::fixed) << setprecision(5)
	       << setw(9) << 0.5
	       << setw(9) << 0.5
	       << resetiosflags(ios::fixed)
	       << " << GENE FREQUENCIES" << endl;

	// Third line
	//  - number of liability classes
	// following: penetrances
	stream << setw(2) << 1
	       << " << NO. OF LIABILITY CLASSES"
	       << endl;

	stream << setiosflags(ios::fixed) << setprecision(4)
	       << setw(8) << 0.0
	       << setw(8) << 0.0
	       << setw(8) << 1.0
	       << resetiosflags(ios::fixed)
	       << " << PENETRANCES"
	       << endl;

	return stream;
}

static ostream& datQuantitative(ostream& stream, const struct Locus *loc) {

	// DUMMY FUNCTION - NO REAL DATA!

	// first line:
	//  - locus type		0, quantitative
	//  - number of alleles		2, low- and high-allel
	//  - locus name		none
	//  - additional comments
	stream << 0
	       << setw(4) << 2
	       << " " << ""
	       << " # qtl << QUANTITATIVE, NO. OF ALLELES"
	       << endl;

	// Second line:
	//  - gene frequencies
	stream << setiosflags(ios::fixed) << setprecision(5)
	       << setw(9) << 0.5
	       << setw(9) << 0.5
	       << resetiosflags(ios::fixed)
	       << " << GENE FREQUENCIES" << endl;

	// Third line and following:
	//  - number of traits				default: 1.0
	//  - genotype means				default: 0.0 0.0 0.0
	//  - variance-covariance matrix		default: 1.0
	//  - multiplier for variance in heterozygotes	default: 1.0
	stream << setw(2) << 1
	       << " << NO. OF TRAITS"
	       << endl;

	stream << setiosflags(ios::fixed | ios::showpoint) << setprecision(3)
	       << setw(10) << 0.0
	       << setw(10) << 0.0
	       << setw(10) << 0.0
	       << " << GENOTYPE MEANS"
	       << resetiosflags(ios::fixed | ios::showpoint)
	       << endl;

	stream << setiosflags(ios::fixed | ios::showpoint) << setprecision(3)
	       << setw(10) << 1.0
	       << " << VARIANCE - COVARIANCE MATRIX"
               << resetiosflags(ios::fixed | ios::showpoint)
               << endl;

	stream << setiosflags(ios::fixed | ios::showpoint) << setprecision(3)
	       << setw(10) << 1.0
	       << " << MULTIPLIER FOR VARIANCE IN HETEROZYGOTES"
               << resetiosflags(ios::fixed | ios::showpoint)
               << endl;

	return stream;
}


ostream& linkage_dat(ostream& stream, bool ignoreAffectionStatus = false, bool ignoreTrait = false) {

	// First line:
	//  - number of loci		number of records in 'db.db_locus'
	//				(+1 affection status)
	//				(+ qtl)
	//  - risk locus		default: 0
	//  - sexlinked			default: 0 (false)
	//  - program			default: 5 (mlink)
	//  - additional comments

	unsigned int nLoci = db.db_locus.size()
			     + (ignoreAffectionStatus ? 0 : 1)
			     + (ignoreTrait ? 0 : 1);

	stream << " " << nLoci
	       << " " << 0
	       << " " << 0
	       << " " << 5
	       << " << NO. OF LOCI, RISK LOCUS, SEXLINKED (IF 1), PROGRAM"
	       << endl;

	// Second line:
	//  - mutation locus		default: 0
	//  - mutation frequency male	default: 0.0 (none)
	//  - mutation frequency female	default: 0.0 (none)
	//  - haplotype frequencies	default: 0   (false)
	//  - additional comments
	stream << " " << 0
	       << setiosflags(ios::showpoint) << setprecision(2)
	       << " " << 0.0
	       << " " << 0.0
	       << resetiosflags(ios::showpoint)
	       << " " << 0
	       << " << MUT LOCUS, MUT MALE, MUT FEM, HAP FREQ (IF 1)"
	       << endl;

	// Third line:
	//  - locus order		as many ascending integers as loci
	//				(+ affection status)
	//				(+ qtl)
	for (unsigned int i = 1; i <= nLoci; i++)
		stream << setw(3) << i;
	stream << endl;

	// first of all, the affection status. Just a dummy!
	if (! ignoreAffectionStatus)
		datAffectionStatus(stream, NULL);

	//
	// Following:
	// loci in ascending order (by locus position)
	//
	LocusTable::iterator lit = db.db_locus.begin();
	while (lit != db.db_locus.end())
		datAllelNumbers(stream, (*lit++).second);

	// Add quantitative locus. Again, a dummy!
	if (! ignoreTrait)
		datQuantitative(stream, NULL);

	// 3 lines from EOF:
	//  - sex difference		default: 0
	//  - interference		default: 0
	stream << setw(2) << 0
	       << setw(2) << 0
	       << " << SEX DIFFERENCE, INTERFERENCE (IF 1 OR 2)"
	       << endl;

	// line before the last line:
	//  - recombination values (theta)
	stream << setiosflags(ios::fixed | ios::showpoint) << setprecision(4);

	// add dummy affection status (!)
	struct Locus affection = { string(""), 0 };
	struct Locus *m, *n;

	// re-use locus iterator from above
	lit = db.db_locus.begin();

	//
	// if(db.db_locus.empty())
	//	lit.begin() == lit.end()
	//  -> infinity loop
	//
	if (lit != db.db_locus.end()) {
		if (! ignoreAffectionStatus) {
			m = &affection;
		} else {
			m = lit->second;
			++lit;
		}
	}

	for ( ; lit != db.db_locus.end(); ++lit) {
		n = lit->second;
		stream << setw(8) << mapfunc(m->loc_pos, n->loc_pos);
		m = n;
	}

	// Add qtl recombination value (0.5 per default)
	// Andre, 2003/10/07: set to '0.1' due to genehunter limitations
	if (! ignoreTrait)
		stream << setw(8) << 0.1;

	stream << " << RECOMBINATION VALUES"
	       << resetiosflags(ios::fixed | ios::showpoint)
	       << endl;

	// last line
	//  - rec varied (?)		default: 1
	//  - increment			default: 0.1
	//  - finishing value		default: 0.45
	stream << " " << 1
	       << setiosflags(ios::fixed | ios::showpoint) << setprecision(5)
	       << " " << 0.1
	       << " " << 0.45
	       << resetiosflags(ios::fixed | ios::showpoint)
	       << " << REC VARIED, INCREMENT, FINISHING VALUE"
	       << endl;

	return stream;
}

